private import SDL;
private import SDL_image;
private import opengl;
private import mymath;
private import std.cstream;
private import std.c.stdio;
private import std.c.stdlib;
private import std.string;
private import std.file;

private import log;

class Texture {
	private uint id_;
	private float texcoord_[4];

    static void init() {
        foreach (char[] f; listdir(`img`)) {
            if (f == `.`) continue;
            if (f == `..`) continue;
            Texture t = Texture.get(`img/` ~ f);
        }
    }

    static Texture[char[]] caches;
    static Texture get(char[] filename) {
        if (!(filename in caches)) {
            caches[filename] = new Texture(filename);
        }
        return caches[filename];
    }
    static Texture get() {
        return caches[caches.keys[rnd(caches.length)]];
    }

    /* Quick utility function for texture creation */
	int power_of_two(int input) {
		int value = 1;

		while ( value < input ) {
			value <<= 1;
		}
		return value;
	}

	this(char[] filename) {
		Log.info(filename ~ ": loading texture");

		SDL_Surface* surf = IMG_Load(filename);
        SDL_SetColorKey(surf, SDL_SRCCOLORKEY, 0);
		this(surf);
		SDL_FreeSurface(surf);
	}

	this(SDL_Surface *surface) {
/+
		/* Use the surface width and height expanded to powers of 2 */
		int w = power_of_two(surface.w);
		int h = power_of_two(surface.h);
		texcoord_[0] = 0.0f;			/* Min X */
		texcoord_[1] = 0.0f;			/* Min Y */
		texcoord_[2] = cast(GLfloat)surface.w / w;	/* Max X */
		texcoord_[3] = cast(GLfloat)surface.h / h;	/* Max Y */

        // from NeHe19
        Uint8* rowhi, rowlo;
        Uint8* tmpbuf;
        Uint8 tmpch;
        SDL_Surface *image;
        int i, j;

        image = surface;
        if ( image == null ) {
            fprintf(stderr, "Unable to load texture: %s\n", SDL_GetError());
            throw new Error("error");
        }

        /* GL surfaces are upsidedown and RGB, not BGR :-) */
        tmpbuf = cast(Uint8 *)malloc(image.pitch);
        if ( tmpbuf == null ) {
            fprintf(stderr, "Out of memory\n");
            throw new Error("error");
        }
        rowhi = cast(Uint8 *)image.pixels;
        rowlo = rowhi + (image.h * image.pitch) - image.pitch;
        for ( i=0; i<image.h/2; ++i ) {
            for ( j=0; j<image.w; ++j ) {
                tmpch = rowhi[j*3];
                rowhi[j*3] = rowhi[j*3+2];
                rowhi[j*3+2] = tmpch;
                tmpch = rowlo[j*3];
                rowlo[j*3] = rowlo[j*3+2];
                rowlo[j*3+2] = tmpch;
            }
            memcpy(tmpbuf, rowhi, image.pitch);
            memcpy(rowhi, rowlo, image.pitch);
            memcpy(rowlo, tmpbuf, image.pitch);
            rowhi += image.pitch;
            rowlo -= image.pitch;
        }
        free(tmpbuf);

		GLuint texture;

		/* Create an OpenGL texture for the image */
		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 3,
					 image.w, image.h,
					 0,
					 GL_RGB,
					 GL_UNSIGNED_BYTE,
					 image.pixels);

		id_ = texture;
+/

		GLuint texture;
		int w, h;
		SDL_Surface *image;
		SDL_Rect area;
		Uint32 saved_flags;
		Uint8  saved_alpha;

		/* Use the surface width and height expanded to powers of 2 */
		w = power_of_two(surface.w);
		h = power_of_two(surface.h);
		texcoord_[0] = 0.0f;			/* Min X */
		texcoord_[1] = 0.0f;			/* Min Y */
		texcoord_[2] = cast(GLfloat)surface.w / w;	/* Max X */
		texcoord_[3] = cast(GLfloat)surface.h / h;	/* Max Y */

		version(LittleEndian) {
			image = SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				w, h,
				32,
				0x000000FF, 
				0x0000FF00, 
				0x00FF0000, 
				0xFF000000);
		}
		version(BigEndian) {
			image = SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				w, h,
				32,
				0xFF000000,
				0x00FF0000, 
				0x0000FF00, 
				0x000000FF);
		}

		if ( image == null ) {
			Log.error("cannot load texture");
		}

		/* Save the alpha blending attributes */
		saved_flags = surface.flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
		saved_alpha = surface.format.alpha;
		if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
			SDL_SetAlpha(surface, 0, 0);
		}

		/* Copy the surface into the GL texture image */
		area.x = 0;
		area.y = 0;
		area.w = surface.w;
		area.h = surface.h;
		SDL_BlitSurface(surface, &area, image, &area);

		/* Restore the alpha blending attributes */
		if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
			SDL_SetAlpha(surface, saved_flags, saved_alpha);
		}

		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,
					 0,
					 3,
					 image.w, image.h,
					 0,
					 GL_RGBA,
					 GL_UNSIGNED_BYTE,
					 image.pixels);

		SDL_FreeSurface(image); /* No longer needed */

		id_ = texture;
	}

	~this() {
		deleteTexture();
	}

	void bind() {
		glBindTexture(GL_TEXTURE_2D, id_);
	}

	GLfloat[] texcoord() {
		return texcoord_;
	}

	void deleteTexture() {
		glDeleteTextures(1, &id_);
	}

}

